@@ -35,6 +35,7 @@
 #include "closeout.h"
 #include "argmatch.h"
 #include "error.h"
+#include "human.h"
 #include "safe-read.h"
 #include "xstrtol.h"
 
@@ -242,52 +243,65 @@ Print the last %d lines of each FILE to standard output.\n\
 With more than one FILE, precede each with a header giving the file name.\n\
 With no FILE, or when FILE is -, read standard input.\n\
 \n\
+Mandatory arguments to long options are mandatory for short options too.\n\
       --retry              keep trying to open a file even if it is\n\
                            inaccessible when tail starts or if it becomes\n\
                            inaccessible later -- useful only with -f\n\
+"), DEFAULT_N_LINES);
+     fputs (_("\
   -c, --bytes=N            output the last N bytes\n\
   -f, --follow[={name|descriptor}]\n\
                            output appended data as the file grows;\n\
                            -f, --follow, and --follow=descriptor are\n\
                            equivalent\n\
   -F                       same as --follow=name --retry\n\
+"), stdout);
+     printf (_("\
   -n, --lines=N            output the last N lines, instead of the last %d\n\
       --max-unchanged-stats=N\n\
                            with --follow=name, reopen a FILE which has not\n\
                            changed size after N (default %d) iterations\n\
                            to see if it has been unlinked or renamed\n\
                            (this is the usual case of rotated log files)\n\
+"),
+	     DEFAULT_N_LINES,
+	     DEFAULT_MAX_N_UNCHANGED_STATS_BETWEEN_OPENS
+	     );
+     fputs (_("\
       --pid=PID            with -f, terminate after process ID, PID dies\n\
   -q, --quiet, --silent    never output headers giving file names\n\
   -s, --sleep-interval=S   with -f, each iteration lasts approximately S\n\
                            (default 1) seconds\n\
+"), stdout);
+     fputs (_("\
   -v, --verbose            always output headers giving file names\n\
       --help               display this help and exit\n\
       --version            output version information and exit\n\
 \n\
-"),
-	      DEFAULT_N_LINES, DEFAULT_N_LINES,
-	      DEFAULT_MAX_N_UNCHANGED_STATS_BETWEEN_OPENS
-	      );
-      printf (_("\
+"), stdout);
+     fputs (_("\
 If the first character of N (the number of bytes or lines) is a `+',\n\
 print beginning with the Nth item from the start of each file, otherwise,\n\
 print the last N items in the file.  N may have a multiplier suffix:\n\
 b for 512, k for 1024, m for 1048576 (1 Meg).  A first OPTION of -VALUE\n\
 or +VALUE is treated like -n VALUE or -n +VALUE unless VALUE has one of\n\
 the [bkm] suffix multipliers, in which case it is treated like -c VALUE\n\
+"), stdout);
+     fputs (_("\
 or -c +VALUE.  Warning: a first option of +VALUE is obsolescent, and support\n\
 for it will be withdrawn.\n\
 \n\
 With --follow (-f), tail defaults to following the file descriptor, which\n\
 means that even if a tail'ed file is renamed, tail will continue to track\n\
 its end.  This default behavior is not desirable when you really want to\n\
+"), stdout);
+     fputs (_("\
 track the actual name of the file, not the file descriptor (e.g., log\n\
 rotation).  Use --follow=name in that case.  That causes tail to track the\n\
 named file by reopening it periodically to see if it has been removed and\n\
 recreated by some other program.\n\
 \n\
-"));
+"), stdout);
       puts (_("\nReport bugs to <bug-textutils@gnu.org>."));
     }
   exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
@@ -310,7 +324,6 @@ static void
 xwrite (int fd, char *const buffer, size_t n_bytes)
 {
   assert (fd == STDOUT_FILENO);
-  assert (n_bytes >= 0);
   if (n_bytes > 0 && fwrite (buffer, 1, n_bytes, stdout) == 0)
     error (EXIT_FAILURE, errno, _("write error"));
 }
@@ -362,21 +375,61 @@ dump_remainder (const char *pretty_filename, int fd, off_t n_bytes)
   return n_written;
 }
 
+/* Call lseek with the specified arguments, where file descriptor FD
+   corresponds to the file, FILENAME.
+   Give a diagnostic and exit nonzero if lseek fails.  */
+
+static void
+xlseek (int fd, off_t offset, int whence, char const *filename)
+{
+  off_t new_offset = lseek (fd, offset, whence);
+  char buf[LONGEST_HUMAN_READABLE + 1];
+  char *s;
+  char const *sign;
+
+  if (0 <= new_offset)
+    return;
+
+  sign = offset < 0 ? "-" : "";
+  if (offset < 0)
+    offset = -offset;
+
+  s = human_readable ((uintmax_t) offset, buf, 1, 1);
+  switch (whence)
+    {
+    case SEEK_SET:
+      error (1, errno, _("%s: cannot seek to offset %s%s"),
+	     filename, sign, s);
+      break;
+    case SEEK_CUR:
+      error (1, errno, _("%s: cannot seek to relative offset %s%s"),
+	     filename, sign, s);
+      break;
+    case SEEK_END:
+      error (1, errno, _("%s: cannot seek to end-relative offset %s%s"),
+	     filename, sign, s);
+      break;
+    default:
+      abort ();
+    }
+}
+
 /* Print the last N_LINES lines from the end of file FD.
    Go backward through the file, reading `BUFSIZ' bytes at a time (except
    probably the first), until we hit the start of the file or have
    read NUMBER newlines.
+   START_POS is the starting position of the read pointer for the file
+   associated with FD (may be nonzero).
    FILE_LENGTH is the length of the file (one more than the offset of the
    last byte of the file).
    Return 0 if successful, 1 if an error occurred.  */
 
 static int
 file_lines (const char *pretty_filename, int fd, long int n_lines,
-	    off_t file_length)
+	    off_t start_pos, off_t file_length)
 {
   char buffer[BUFSIZ];
   int bytes_read;
-  int i;			/* Index into `buffer' for scanning.  */
   off_t pos = file_length;
 
   if (n_lines == 0)
@@ -384,14 +437,13 @@ file_lines (const char *pretty_filename, int fd, long int n_lines,
 
   /* Set `bytes_read' to the size of the last, probably partial, buffer;
      0 < `bytes_read' <= `BUFSIZ'.  */
-  bytes_read = pos % BUFSIZ;
+  bytes_read = (pos - start_pos) % BUFSIZ;
   if (bytes_read == 0)
     bytes_read = BUFSIZ;
   /* Make `pos' a multiple of `BUFSIZ' (0 if the file is short), so that all
      reads will be on block boundaries, which might increase efficiency.  */
   pos -= bytes_read;
-  /* FIXME: check lseek return value  */
-  lseek (fd, pos, SEEK_SET);
+  xlseek (fd, pos, SEEK_SET, pretty_filename);
   bytes_read = safe_read (fd, buffer, bytes_read);
   if (bytes_read == -1)
     {
@@ -405,6 +457,7 @@ file_lines (const char *pretty_filename, int fd, long int n_lines,
 
   do
     {
+      int i;			/* Index into `buffer' for scanning.  */
       /* Scan backward, counting the newlines in this bufferfull.  */
       for (i = bytes_read - 1; i >= 0; i--)
 	{
@@ -421,17 +474,15 @@ file_lines (const char *pretty_filename, int fd, long int n_lines,
 	    }
 	}
       /* Not enough newlines in that bufferfull.  */
-      if (pos == 0)
+      if (pos == start_pos)
 	{
 	  /* Not enough lines in the file; print the entire file.  */
-	  /* FIXME: check lseek return value  */
-	  lseek (fd, (off_t) 0, SEEK_SET);
+	  xlseek (fd, start_pos, SEEK_SET, pretty_filename);
 	  dump_remainder (pretty_filename, fd, file_length);
 	  return 0;
 	}
       pos -= BUFSIZ;
-      /* FIXME: check lseek return value  */
-      lseek (fd, pos, SEEK_SET);
+      xlseek (fd, pos, SEEK_SET, pretty_filename);
     }
   while ((bytes_read = safe_read (fd, buffer, BUFSIZ)) > 0);
 
@@ -836,8 +887,7 @@ recheck (struct File_spec *f)
       f->n_unchanged_stats = 0;
       f->n_consecutive_size_changes = 0;
       f->ignore = 0;
-      /* FIXME: check lseek return value  */
-      lseek (f->fd, f->size, SEEK_SET);
+      xlseek (f->fd, f->size, SEEK_SET, pretty_name (f));
     }
 }
 
@@ -937,8 +987,8 @@ tail_forever (struct File_spec *f, int nfiles)
 	    {
 	      error (0, 0, _("%s: file truncated"), pretty_name (&f[i]));
 	      last = i;
-	      /* FIXME: check lseek return value  */
-	      lseek (f[i].fd, stats.st_size, SEEK_SET);
+	      xlseek (f[i].fd, (off_t) stats.st_size, SEEK_SET,
+		      pretty_name (&f[i]));
 	      f[i].size = stats.st_size;
 	      continue;
 	    }
@@ -1000,8 +1050,7 @@ tail_bytes (const char *pretty_filename, int fd, off_t n_bytes)
     {
       if (S_ISREG (stats.st_mode))
 	{
-	  /* FIXME: check lseek return value  */
-	  lseek (fd, n_bytes, SEEK_CUR);
+	  xlseek (fd, n_bytes, SEEK_CUR, pretty_filename);
 	}
       else if (start_bytes (pretty_filename, fd, n_bytes))
 	{
@@ -1036,15 +1085,13 @@ tail_bytes (const char *pretty_filename, int fd, off_t n_bytes)
 		 more bytes than have been requested.  So reposition the
 		 file pointer to the incoming current position and print
 		 everything after that.  */
-	      /* FIXME: check lseek return value  */
-	      lseek (fd, current_pos, SEEK_SET);
+	      xlseek (fd, current_pos, SEEK_SET, pretty_filename);
 	    }
 	  else
 	    {
 	      /* There are more bytes remaining than were requested.
 		 Back up.  */
-	      /* FIXME: check lseek return value  */
-	      lseek (fd, -n_bytes, SEEK_END);
+	      xlseek (fd, -n_bytes, SEEK_END, pretty_filename);
 	    }
 	  dump_remainder (pretty_filename, fd, n_bytes);
 	}
@@ -1061,7 +1108,6 @@ static int
 tail_lines (const char *pretty_filename, int fd, long int n_lines)
 {
   struct stat stats;
-  off_t length;
 
   /* We need binary input, since `tail' relies on `lseek' and byte counts,
      while binary output will preserve the style (Unix/DOS) of text file.  */
@@ -1081,17 +1127,17 @@ tail_lines (const char *pretty_filename, int fd, long int n_lines)
     }
   else
     {
-      /* Use file_lines only if FD refers to a regular file with
-         its file pointer positioned at beginning of file.  */
-      /* FIXME: adding the lseek conjunct is a kludge.
-	 Once there's a reasonable test suite, fix the true culprit:
-	 file_lines.  file_lines shouldn't presume that the input
-	 file pointer is initially positioned to beginning of file.  */
+      off_t length;
+      off_t start_pos;
+
+      /* Use file_lines only if FD refers to a regular file for
+	 which lseek (... SEEK_END) works.  */
       if (S_ISREG (stats.st_mode)
-	  && lseek (fd, (off_t) 0, SEEK_CUR) == (off_t) 0)
+	  && (start_pos = lseek (fd, (off_t) 0, SEEK_CUR)) != -1
+	  && start_pos < (length = lseek (fd, (off_t) 0, SEEK_END)))
 	{
-	  length = lseek (fd, (off_t) 0, SEEK_END);
-	  if (length != 0 && file_lines (pretty_filename, fd, n_lines, length))
+	  if (length != 0 && file_lines (pretty_filename, fd, n_lines,
+					 start_pos, length))
 	    return 1;
 	}
       else
@@ -1120,7 +1166,6 @@ static int
 tail_file (struct File_spec *f, off_t n_units)
 {
   int fd, errors;
-  struct stat stats;
 
   int is_stdin = (STREQ (f->name, "-"));
 
@@ -1156,8 +1201,9 @@ tail_file (struct File_spec *f, off_t n_units)
       errors = tail (pretty_name (f), fd, n_units);
       if (forever)
 	{
+	  struct stat stats;
+
 	  f->errnum = 0;
-	  /* FIXME: duplicate code */
 	  if (fstat (fd, &stats) < 0)
 	    {
 	      errors = 1;
